// Copyright (c) 2025 shibedrill
// SPDX-License-Identifier: GPL-3.0-or-later

#![no_std]
#![no_main]
#![feature(cfg_match)]
#![feature(allocator_api)]

mod arch;
mod boot;
mod constants;
#[macro_use]
mod log;
mod memory;
mod panic;
mod params;
mod process;
mod resources;
mod syscall_runner;

use core::ptr;

use arch::x86_64::gdt::{self, GDT};
use arch::x86_64::serial::Serialport;
use boot::*;
use constants::*;
use log::*;
use memory::alloc::{format, string::*, vec};
use params::*;

use limine::file::File;
use limine::memory_map::EntryType;
#[allow(unused_imports)]
use lzma_rs::lzma_decompress;

#[unsafe(no_mangle)]
unsafe extern "C" fn main() -> ! {
    // Assert supported bootloader version
    assert!(BASE_REVISION.is_supported());

    // TODO: This is hacky and will not work on any other machine.
    let qemu_serial = Serialport::new(0x3f8);

    let executable_file_response = FILE_REQUEST
        .get_response()
        .expect("Bootloader did not return executable data");
    let params =
        get_kernel_params(limine::file::File::string(executable_file_response.file()).to_bytes());

    // Set up logging level from params
    // Nothing we can do here if this fails since no log subscribers are initialized yet
    if let Some(level) = params.get("-loglevel")
        && let Ok(parsed_level) = LogLevel::try_from(level.as_str())
    {
        LOGGER.set_level(parsed_level);
    }
    // Add subscribers to logger
    if let Some(device) = params.get("-logdev") {
        let log_device_list: vec::Vec<&str> = device.split(',').collect();
        if log_device_list.contains(&"display") {
            // Append display console to log subs
        }
        if log_device_list.contains(&"serial") {
            LOGGER.add_subscriber(qemu_serial);
        }
        log_trace!("Boot: Configured kernel logging devices");
    }
    log_info!(
        "Boot: Kernel cmdline: {}",
        String::from_utf8_lossy(
            limine::file::File::string(executable_file_response.file()).to_bytes()
        )
    );
    log_info!(
        "Boot: Kernel file path: {}",
        String::from_utf8_lossy(executable_file_response.file().path().to_bytes())
    );
    log_info!("Boot: Booting gila version {}", kernel_version_string());

    log_info!("Boot: Trans rights!");

    if let Some(framebuffer_response) = FRAMEBUFFER_REQUEST.get_response() {
        let fb = framebuffer_response.framebuffers().next().unwrap();
        log_info!("Boot: Framebuffer response received");
        log_trace!(
            "Boot: Framebuffer dimensions: {}x{}, {} bits per pixel",
            fb.width(),
            fb.height(),
            fb.bpp()
        );
    } else {
        log_info!("Boot: Framebuffer response absent, graphics disabled",);
    }

    let smp_response = MP_REQUEST.get_response();
    match smp_response {
        None => log_info!("MP: Multiprocessing response not received"),
        Some(resp) => {
            log_info!("MP: Multiprocessing response received");
            log_trace!("MP: {} CPUs found", resp.cpus().len());
        }
    }

    let _rsdp_response = RSDP_REQUEST.get_response();

    let module_response = MODULE_REQUEST
        .get_response()
        .expect("Bootloader did not return kernel modules");
    log_info!(
        "Boot: {} kernel modules found",
        module_response.modules().len()
    );
    if !module_response.modules().is_empty() {
        let mut log_msg: String = String::from("Boot: Kernel modules list:\n");
        for module in module_response.modules() {
            log_msg.push_str(&format!(
                "\t\t{}",
                String::from_utf8_lossy(module.path().to_bytes())
            ));
        }
        log_trace!("{log_msg}")
    }

    let irfs_path = {
        if let Some(key) = params.get("-initramfs") {
            key
        } else {
            INITRAMFS_DEFAULT_PATH
        }
    };

    log_trace!("Boot: initramfs path requested: {irfs_path}");
    let _initramfs = {
        let mut result: Option<&File> = None;
        for file in module_response.modules() {
            if file.path().to_string_lossy() == irfs_path {
                result = Some(file);
                log_info!("Boot: initramfs found");
            }
        }
        result
    }
    .expect("initramfs not found in modules list");

    if let Some(mmap_response) = MEMMAP_REQUEST.get_response() {
        log_info!("Boot: Memory map received");
        if !mmap_response.entries().is_empty() {
            let mut log_msg: String = String::from("Boot: Memory map:");
            let mut usable: u64 = 0;
            let mut reclaimable: u64 = 0;
            let mut hardware: u64 = 0;
            let mut unusable: u64 = 0;
            for entry in mmap_response.entries() {
                log_msg.push_str(&format!(
                    "\n\t\t0x{:x} bytes @ 0x{:x}: {}",
                    entry.length,
                    entry.base,
                    match entry.entry_type {
                        EntryType::ACPI_NVS => {
                            hardware += entry.length;
                            "ACPI (reserved)"
                        }
                        EntryType::ACPI_RECLAIMABLE => {
                            reclaimable += entry.length;
                            "ACPI (reclaimable)"
                        }
                        EntryType::BAD_MEMORY => {
                            unusable += entry.length;
                            "damaged/unusable"
                        }
                        EntryType::BOOTLOADER_RECLAIMABLE => {
                            reclaimable += entry.length;
                            "bootloader (reclaimable)"
                        }
                        #[allow(unreachable_patterns, deprecated)]
                        EntryType::EXECUTABLE_AND_MODULES | EntryType::KERNEL_AND_MODULES => {
                            unusable += entry.length;
                            "executable & modules"
                        }
                        EntryType::FRAMEBUFFER => {
                            hardware += entry.length;
                            "framebuffer"
                        }
                        EntryType::RESERVED => {
                            unusable += entry.length;
                            "reserved"
                        }
                        EntryType::USABLE => {
                            usable += entry.length;
                            "usable"
                        }
                        _ => "unidentified",
                    }
                ));
            }
            log_trace!("{log_msg}");
            let total = usable + reclaimable + hardware + unusable;
            log_info!(
                "Boot: Memory report:\n\t\tFree: {usable}\n\t\tAvailable: {reclaimable}\n\t\tUsable: {}\n\t\tHardware: {hardware}\n\t\tUnusable: {unusable}\n\t\tTotal: {}",
                usable + reclaimable,
                total
            );
        }
    }

    log_trace!(
        "GDT: Setting up global descriptor table at 0x{:x}",
        ptr::addr_of!(GDT) as usize
    );
    gdt::setup_gdt();

    #[allow(unreachable_code)]
    loop {
        for _i in 0..50000000 {
            unsafe {
                core::arch::asm!("nop");
            }
        }
        log_trace!("Heartbeat");
    }
}
